;Word search
;It works by doing a binary search on a sorted list of words.
;On my 1.6 GHz PIV it does 1 545 000 searches in just 0.65 sec,
;fast enough for my needs.
;It uses an average of 9.66 comparitios to find a word when the wordlist is 1545 words.
;
; KetilO (C) 2003

.386
.model flat, stdcall  ;32 bit memory model
option casemap :none  ;case sensitive

include FindWord.inc

.code

start:

	invoke GetModuleHandle,NULL
	mov		hInstance,eax

    invoke InitCommonControls
	invoke DialogBoxParam,hInstance,IDD_DIALOG1,NULL,addr DlgProc,NULL
	invoke ExitProcess,0

;########################################################################

;Sort the words
SortWords proc uses ebx esi edi

	mov		ebx,hMemPtrs
	xor		ecx,ecx
	.while ecx<nWordCount
		mov		edx,ecx
		inc		edx
		.while edx<nWordCount
			mov		esi,[ebx+ecx*4]
			mov		edi,[ebx+edx*4]
			dec		esi
			dec		edi
		  @@:
			inc		esi
			inc		edi
			mov		al,[edi]
			cmp		al,[esi]
			jne		@f
			or		al,al
			jne		@b
		  @@:
			jnc		@f
			mov		eax,[ebx+ecx*4]
			xchg	eax,[ebx+edx*4]
			mov		[ebx+ecx*4],eax
		  @@:
			inc		edx
		.endw
		inc		ecx
	.endw
	ret

SortWords endp

;Set up the word pointer array
SetupWordPointers proc uses esi edi

	mov		eax,nWordCount
	.if eax
		;Allocate memory for word pointers
		lea		eax,[eax*4+4]
		invoke GlobalAlloc,GMEM_FIXED or GMEM_ZEROINIT,eax
		mov		hMemPtrs,eax
		mov		esi,hMemWords
		mov		edi,hMemPtrs
		;Setup word pointers
		.while byte ptr [esi]
			mov		[edi],esi
			add		edi,4
			.while byte ptr [esi]
				inc		esi
			.endw
			inc		esi
		.endw
	.endif
	ret

SetupWordPointers endp

;Parse the words from CR+LF terminated to ZERO terminated
ParseWords proc

	mov		nWordCount,0
	mov		esi,hMemWords
	.if esi
		mov		edi,esi
	  Nx:
		mov		al,[esi]
		cmp		al,0Dh
		je		@f
		mov		[edi],al
		inc		esi
		inc		edi
		or		al,al
		je		Ex
		jmp		Nx
	  @@:
		inc		nWordCount
		mov		al,0
		mov		[edi],al
		inc		edi
		add		esi,2
		mov		al,[esi]
		or		al,al
		jne		Nx
	  Ex:
		mov		[edi],al
		xor		eax,eax
	.endif
	ret

ParseWords endp

;Read the words from file
ReadWordFile proc
	LOCAL	hFile:DWORD
	LOCAL	nBytes:DWORD

	invoke CreateFile,offset szWordFile,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0
	.if eax!=INVALID_HANDLE_VALUE
		mov		hFile,eax
		invoke GetFileSize,hFile,addr nBytes
		inc		eax
		mov		nBytes,eax
		invoke GlobalAlloc,GMEM_FIXED or GMEM_ZEROINIT,eax
		mov		hMemWords,eax
		invoke ReadFile,hFile,hMemWords,nBytes,addr nBytes,FALSE
		invoke CloseHandle,hFile
	.endif
	ret

ReadWordFile endp

;Find the word using successive approximation
FindWord proc uses ebx esi edi,lpWord:DWORD
	LOCAL	nLU:DWORD
	LOCAL	nLD:DWORD

	xor		eax,eax
	mov		nLD,eax
	mov		ebx,hMemPtrs
	mov		ecx,nWordCount
	dec		ecx
	mov		nLU,ecx
	shr		ecx,1
	mov		esi,lpWord
  Nxt:
	inc		nScans
	mov		edi,[ebx+ecx*4]
	xor		edx,edx
  @@:
	mov		al,[edi+edx]
	cmp		al,[esi+edx]
	jne		@f
	inc		edx
	or		al,al
	jne		@b
	;Word found, return index
	mov		eax,ecx
	jmp		Ex
  @@:
	jnc		Dn
	;Up
	mov		eax,nLU
	sub		eax,ecx
	shr		eax,1
	.if !eax
		cmp		ecx,nLD
		je		Nf
		inc		eax
	.endif
	mov		nLD,ecx
	add		ecx,eax
	cmp		ecx,nWordCount
	jc		Nxt
	jmp		Nf
  Dn:
	;Down
	mov		eax,ecx
	sub		eax,nLD
	shr		eax,1
	.if !eax
		cmp		ecx,nLU
		je		Nf
		inc		eax
	.endif
	mov		nLU,ecx
	sub		ecx,eax
	jnc		Nxt
  Nf:
	;Word not found, return -1
	xor		eax,eax
	dec		eax
  Ex:
	ret

FindWord endp

DlgProc proc hWin:HWND,uMsg:UINT,wParam:WPARAM,lParam:LPARAM
	LOCAL	nNotFound:DWORD
	LOCAL	buffer[256]:BYTE

	mov		eax,uMsg
	.if eax==WM_INITDIALOG
		invoke ReadWordFile
		invoke ParseWords
		invoke SetupWordPointers
		invoke SortWords
		invoke SetDlgItemInt,hWin,1001,nWordCount,FALSE
		push	ebx
		mov		ebx,hMemPtrs
		xor		ecx,ecx
		.while dword ptr [ebx+ecx*4]
			push	ecx
			invoke SendDlgItemMessage,hWin,1002,LB_ADDSTRING,0,[ebx+ecx*4]
			pop		ecx
			inc		ecx
		.endw
		pop		ebx
	.elseif eax==WM_COMMAND
		mov		eax,wParam
		mov		edx,eax
		shr		edx,16
		and		eax,0FFFFh
		.if edx==BN_CLICKED
			.if eax==IDOK
				;Do nWordCount * 1000 searches to test speed
;invoke GetTickCount
;push	eax
				mov		ecx,1000
			  Nxt:
				push	ecx
				;Test if it works by finding all the words
				xor		ecx,ecx
				mov		nNotFound,ecx
				mov		nScans,ecx
				.while ecx<nWordCount
					push	ecx
					mov		eax,hMemPtrs
					lea		eax,[eax+ecx*4]
					mov		eax,[eax]
					invoke FindWord,eax
					.if eax==-1
						;Bug, word not found
						inc		nNotFound
					.endif
					pop		ecx
					inc		ecx
				.endw
				pop		ecx
				loop	Nxt
				.if nNotFound
					;Bug, words not found
					invoke SetDlgItemInt,hWin,1003,nNotFound,FALSE
				.endif
				invoke SetDlgItemInt,hWin,1004,nScans,FALSE
;invoke GetTickCount
;pop		edx
;sub		eax,edx
;PrintDec eax
			.elseif eax==IDCANCEL
				invoke SendMessage,hWin,WM_CLOSE,NULL,NULL
			.endif
		.elseif edx==EN_CHANGE
			.if eax==1003
				;Find a word
				invoke GetDlgItemText,hWin,1003,addr buffer,sizeof buffer
				mov		nScans,0
				invoke FindWord,addr buffer
				invoke SendDlgItemMessage,hWin,1002,LB_SETCURSEL,eax,0
				invoke SetDlgItemInt,hWin,1004,nScans,FALSE
			.endif
		.endif
	.elseif eax==WM_CLOSE
		invoke GlobalFree,hMemWords
		invoke GlobalFree,hMemPtrs
		invoke EndDialog,hWin,0
	.else
		mov		eax,FALSE
		ret
	.endif
	mov		eax,TRUE
	ret

DlgProc endp

end start
